import click

from kks.cmd.build import build
from kks.util.click import GroupedGroup
from kks.util.common import config_directory
from kks.cmd.auth import auth
from kks.cmd.convert import convert
from kks.cmd.deadlines import deadlines
from kks.cmd.gen import gen
from kks.cmd.hide import hide, unhide
from kks.cmd.init import init
from kks.cmd.lint import lint
from kks.cmd.open import open_
from kks.cmd.run import run
from kks.cmd.status import status
from kks.cmd.submit import submit
from kks.cmd.sync import sync
from kks.cmd.test import test_
from kks.cmd.top import top
from kks.cmd.upgrade import upgrade, update


def cleanup():
    for filename in ['cookies.pickle', 'targets.yaml']:
        file = config_directory() / filename
        if file.exists():
            file.unlink()


@click.group(cls=GroupedGroup)
def cli():
    """KoKoS helper tool"""
    cleanup()


class Commands:
    workspace = ('Workspace-related', 1)
    ejudge = ('Ejudge interaction', 2)
    solution = ('Solution testing', 3)
    other = ('Other', 4)


cli.add_command(init, group=Commands.workspace)
cli.add_command(sync, group=Commands.workspace)
cli.add_command(hide, group=Commands.workspace)
cli.add_command(unhide, group=Commands.workspace)

cli.add_command(auth, group=Commands.ejudge)
cli.add_command(open_, group=Commands.ejudge)
cli.add_command(status, group=Commands.ejudge)
cli.add_command(top, group=Commands.ejudge)
cli.add_command(deadlines, group=Commands.ejudge)
cli.add_command(submit, group=Commands.ejudge)

cli.add_command(lint, group=Commands.solution)
cli.add_command(build, group=Commands.solution)
cli.add_command(run, group=Commands.solution)
cli.add_command(gen, group=Commands.solution)
cli.add_command(test_, group=Commands.solution)

cli.add_command(convert, group=Commands.other)
cli.add_command(upgrade, group=Commands.other)
cli.add_command(update, group=Commands.other)



import click


class Choice2(click.Choice):
    """ for nice help message """
    def get_metavar(self, param):
        if len(self.choices) == 1:
            return self.choices[0]
        return "[{}]".format("|".join(self.choices))


# classes to create a flag with optional value, see https://stackoverflow.com/a/44144098
# click 8.0 should support something like this (https://github.com/pallets/click/issues/549)
class FlagOption(click.Option):
    """ Mark this option as getting a _opt option """
    is_optflag = True


class OptFlagOption(click.Option):
    """ Fix the help for the _opt suffix """
    def get_help_record(self, ctx):
        help = super().get_help_record(ctx)
        return (help[0].replace('_opt ', '='),) + help[1:]

    def get_error_hint(self, ctx):
        hint = super().get_error_hint(ctx)
        return hint.replace('_opt', '')


class OptFlagCommand(click.Command):
    """ Command with support for flags with values """
    def parse_args(self, ctx, args):
        """ Translate any flag= to flag_opt= as needed """
        options = [o for o in ctx.command.params
                   if getattr(o, 'is_optflag', None)]
        prefixes = {p for p in sum([o.opts for o in options], [])
                    if p.startswith('--')}
        for i, a in enumerate(args):
            a = a.split('=')
            if a[0] in prefixes and len(a) > 1:
                a[0] += '_opt'
                args[i] = '='.join(a)

        return super().parse_args(ctx, args)


class GroupedGroup(click.Group):
    def add_command(self, command, *args, **kwargs):
        help_group = kwargs.pop('group', None)
        command.help_group = help_group
        return super().add_command(command, *args, **kwargs)

    def format_commands(self, ctx, formatter):
        """see https://stackoverflow.com/a/58770064"""
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            if not (cmd is None or cmd.hidden):
                commands.append((subcommand, cmd))

        if commands:
            longest = max(len(cmd[0]) for cmd in commands)
            limit = formatter.width - 6 - longest

            groups = {}
            for subcommand, cmd in commands:
                help_str = cmd.get_short_help_str(limit)
                subcommand += ' ' * (longest - len(subcommand))
                groups.setdefault(cmd.help_group, []).append((subcommand, help_str))

            with formatter.section('Commands'):
                for (group_name, _), rows in sorted(groups.items(), key=lambda x: x[0][1]):
                    formatter.write_heading(group_name)
                    with formatter.indentation():
                        formatter.write_dl(rows)
                        formatter.write_paragraph()



import click
from click.exceptions import ClickException


class EjudgeError(ClickException):
    def __init__(self, message='unknown error'):
        super().__init__(message)


class EjudgeUnavailableError(EjudgeError):
    def __init__(self, message='Ejudge is not available'):
        super().__init__(message)


class AuthError(EjudgeError):
    def __init__(self, message='Auth error', fg='red'):
        super().__init__(message)
        self.fg = fg

    def show(self):
        click.secho(self.message, fg=self.fg, err=True)


class APIError(EjudgeError):
    INVALID_RESPONSE = -2
    UNKNOWN = -1

    INVALID_SESSION = 148

    def __init__(self, message, code):
        super().__init__(message)
        self.code = code



from datetime import datetime, timedelta
from itertools import groupby

import click

from kks.ejudge import Deadlines, Status, ejudge_summary, get_contest_deadlines
from kks.util.fancytable import FancyTable, StaticColumn
from kks.util.ejudge import EjudgeSession
from kks.util.storage import Cache, Config


class ContestStatusRow:
    def __init__(self, contest, problem_mapping):
        self.contest = contest.name
        self.penalty = contest.current_penalty
        self.status = 'No deadlines'
        self.deadline = ''
        self._color = contest.deadline_color()
        self._bold = False

        if contest.past_deadline():
            self.status = 'Past deadline'
            self.penalty = '-'
        elif contest.active_deadline() is not None:
            self.status = 'Next deadline'
            self.deadline = contest.deadlines.to_str(contest.active_deadline())
            warn = contest.deadline_is_close()
            self._bold = warn
        if not contest.past_deadline() and all(problem.status in [Status.OK, Status.OK_AUTO, Status.REVIEW] for problem in problem_mapping[contest.name]):
            self._color = 'bright_black'

    def color(self):
        return self._color

    def bold(self):
        return self._bold


@click.command(short_help='Show contest deadlines')
@click.option('-l', '--last', type=int,
              help='Show deadlines for last N contests')
@click.option('-c', '--contest', 'contests', type=str, multiple=True,
              help='Show deadlines for the selected contest')
@click.option('-C', '-nc', '--no-cache', is_flag=True,
              help='Reload cached data')
def deadlines(last, contests, no_cache):
    if last and contests:
        click.secho('"--last" and "--contest" are exclusive, specify no more than one')
        return

    session = EjudgeSession()
    summary = ejudge_summary(session)
    problem_mapping = {contest: list(problems) for contest, problems in groupby(summary, lambda problem: problem.contest())}
    contest_info = get_contest_deadlines(session, summary, no_cache)

    if contests:
        contest_info = [contest for contest in contest_info if contest.name in contests]
    if last:
        contest_info = contest_info[-last:]

    rows = [ContestStatusRow(contest, problem_mapping) for contest in contest_info]

    table = FancyTable()
    table.add_column(StaticColumn('Contest', 4, lambda row: row.contest))
    table.add_column(StaticColumn('Penalty', 3, lambda row: row.penalty))
    table.add_column(StaticColumn.padding(1))
    table.add_column(StaticColumn('Status', 13, lambda row: row.status, right_just=False))
    table.add_column(StaticColumn('Next deadline', len(Deadlines.PLACEHOLDER), lambda row: row.deadline, right_just=False))
    table.show(rows)



import click

from kks.binary import compile_solution
from kks.util.common import get_solution_directory
from kks.util.testing import RunOptions


@click.command(short_help='Build solution')
@click.option('-T', '-tg', '--target', default='default',
              help='Target name to build')
@click.option('-v', '--verbose', is_flag=True,
              help='Verbose mode (show used compiler args)')
@click.option('--asan/--no-asan', is_flag=True, default=None,
              help='Use asan (true by default)')
def build(target, verbose, asan):
    directory = get_solution_directory()

    options = RunOptions(
        asan=asan,
    )

    compile_solution(directory, target, verbose, options)



from os import environ
from pathlib import Path

import click

from kks.ejudge import Links
from kks.util.common import format_file
from kks.util.h2t import HTML2Text
from kks.util.storage import Config


@click.command(short_help='Convert HTML statements to Markdown')
@click.option('-f', '--force', is_flag=True, default=False)
@click.argument('files', nargs=-1, type=click.Path(exists=True))
def convert(force, files):
    """Convert statements from HTML to Markdown (e.g. kr statements copied by scp)"""

    md_width = Config().options.mdwidth
    converter = HTML2Text(bodywidth=md_width, baseurl=Links.WEB_CLIENT_ROOT)  # NOTE baseurl may be incorrect
    converter.pad_tables = True

    for filename in files:
        file = Path(filename)
        md_file = file.with_suffix('.md')

        if not file.is_file():
            click.secho(format_file(file), nl=False)
            click.secho(' is not a file, skipping', fg='red')
            continue
        if md_file.exists():
            if not md_file.is_file():
                click.secho(format_file(md_file), nl=False)
                click.secho(' exists and is not a file, skipping', fg='red')
                continue
            elif not force:
                click.secho(format_file(md_file), nl=False)
                click.secho(' exists, skipping (use --force to overwrite)', fg='yellow')
                continue

        md_file.write_text(converter.handle(file.read_text()))
        click.secho('Converted ', fg='green', nl=False)
        click.secho(format_file(file), nl=False)
        click.secho(' to ', fg='green', nl=False)
        click.secho(format_file(md_file))



import os
import sys

import click
from click._compat import isatty


class Column:
    def header(self):
        raise NotImplementedError()

    def value(self, row):
        raise NotImplementedError()

    def width(self):
        raise NotImplementedError()


class StaticColumn(Column):
    def __init__(self, name, width, mapper, right_just=True):
        self.name = name
        if self.name is None:
            self.name = ''
        self.mapper = mapper
        self.actual_width = max(width, len(self.name))
        self.right_just = right_just

    def _justify(self, value):
        return value.rjust(self.actual_width, ' ') if self.right_just \
            else value.ljust(self.actual_width, ' ')

    def header(self):
        return click.style(self._justify(self.name), fg='white', bold=True)

    def value(self, row):
        return click.style(self._justify(str(self.mapper(row))), fg=row.color(), bold=row.bold())

    def width(self):
        return self.actual_width

    @classmethod
    def padding(cls, width):
        # width - 1, так как лишняя колонка уже добавляет один пробел
        return cls(None, width - 1, lambda _: '')


class DelimiterRow:
    def __init__(self, char='—', color='white', bold=False):
        self.char = char
        self.color = color
        self.bold = bold

    def value(self, width):
        return click.style(self.char * width, fg=self.color, bold=self.bold)


class FancyTable:
    def __init__(self):
        self.columns = []

    def add_column(self, column):
        self.columns.append(column)

    def calc_width(self):
        content = sum([column.width() for column in self.columns])
        return content + len(self.columns) - 1

    def render(self, rows):
        lines = [
            ' '.join([
                column.header()
                for column in self.columns
            ])
        ]

        for row in rows:
            if isinstance(row, DelimiterRow):
                lines.append(row.value(self.calc_width()))
            else:
                lines.append(' '.join([
                    column.value(row)
                    for column in self.columns
                ]))

        return lines

    def show(self, rows, allow_high_tables=False):
        terminal_width, terminal_height = click.get_terminal_size()
        exceeds_width = self.calc_width() > terminal_width
        exceeds_height = len(rows) > terminal_height

        lines = self.render(rows)
        output = '\n'.join(lines)

        if isatty(sys.stdout) and (exceeds_width or exceeds_height and not allow_high_tables):
            if 'LESS' not in os.environ:
                os.environ['LESS'] = '-S -R'
            click.echo_via_pager(output)
        else:
            click.secho(output)
